import pytest
import testing_utils as utils
from testing_utils import logger
from aioinflux.compat import pd, np

if pd is not None:
    pd.set_option('display.max_columns', 10)
    pd.set_option('display.width', 100)


@utils.requires_pandas
def test_write_dataframe(df_client):
    df1 = utils.random_dataframe()
    df2 = utils.random_dataframe()
    df2.columns = df1.columns
    assert df_client.write(df1, measurement='m1', mytag='foo', tag_columns=['tag'])
    assert df_client.write(df2, measurement='m2', mytag='foo', tag_columns=['tag'])
    assert df_client.write(utils.random_dataframe(), measurement='m3')  # tag-less


@utils.requires_pandas
def test_write_dataframe_with_nan(df_client):
    df = utils.trading_df()
    df_client.write(df, f'fills00')
    for i in range(10):
        for _ in range(int(len(df) / 5)):
            i = np.random.randint(df.shape[0])
            j = np.random.randint(df.shape[1])
            df.iloc[i, j] = np.nan
        df_client.write(df, f'fills{i + 1:02d}')


@utils.requires_pandas
def test_select_into(df_client):
    df_client.query("SELECT * INTO m2_copy from m2")
    df = df_client.query('SELECT * from m2_copy')
    assert df.shape == (50, 8)
    logger.info(f'\n{df.head()}')


@utils.requires_pandas
def test_read_dataframe(df_client):
    df = df_client.query('SELECT * from m1')
    logger.info(f'\n{df.head()}')
    assert df.shape == (50, 8)


@utils.requires_pandas
@pytest.mark.asyncio
async def test_dataframe_chunked_query(client):
    client.output = 'dataframe'

    df1 = utils.random_dataframe()
    await client.write(df1, measurement='m3')

    cursor = await client.query('SELECT * FROM m3', chunked=True, chunk_size=10)
    dfs = []
    async for subdf in cursor:
        assert isinstance(subdf, pd.DataFrame)
        assert len(subdf) == 10
        dfs.append(subdf)
    df = pd.concat(dfs)
    assert df.shape == (50, 7)

    client.output = 'json'


@utils.requires_pandas
def test_read_dataframe_groupby(df_client):
    df_dict = df_client.query('SELECT max(*) from /m[1-2]$/ GROUP BY "tag"')
    s = ['\n{}:\n{}'.format(k, v) for k, v in df_dict.items()]
    logger.info('\n'.join(s))
    m1 = pd.concat([df for k, df in df_dict.items() if k.split(',')[0] == 'm1'])
    m2 = pd.concat([df for k, df in df_dict.items() if k.split(',')[0] == 'm2'])
    assert m1.shape == (5, 6)
    assert m2.shape == (5, 6)


@utils.requires_pandas
def test_read_dataframe_multistatement(df_client):
    df_list = df_client.query('SELECT max(*) from m1;SELECT min(*) from m2')
    logger.info(df_list)
    assert type(df_list) is list
    assert 'm1' in df_list[0]
    assert 'm2' in df_list[1]
    assert df_list[0]['m1'].shape == (1, 5)
    assert df_list[1]['m2'].shape == (1, 5)


@utils.requires_pandas
def test_read_dataframe_show_databases(df_client):
    df = df_client.show_databases()
    assert isinstance(df.index, pd.RangeIndex)
    assert 'name' in df.columns
    logger.info(f'\n{df.head()}')


@utils.requires_pandas
@pytest.mark.asyncio
async def test_change_db(client):
    state = client.db, client.output
    client.output = 'dataframe'

    client.db = 'foo'
    await client.ping()

    client.db, client.output = state


###############
# Error tests #
###############

@utils.requires_pandas
@pytest.mark.asyncio
async def test_invalid_data_write_dataframe(client):
    with pytest.raises(ValueError) as e:
        # Non-DatetimeIndex DataFrame
        await client.write(utils.random_dataframe().reset_index(), measurement='foo')
    logger.error(e)

    with pytest.raises(ValueError) as e:
        # DataFrame write without specifying measurement
        await client.write(utils.random_dataframe())
    logger.error(e)


@utils.requires_pandas
def test_chunked_dataframe(df_client):
    with pytest.raises(ValueError) as e:
        _ = df_client.query('SELECT * FROM foo', chunked=True)
    logger.error(e)



import asyncio
import json
import logging
import warnings
from functools import wraps
from typing import TypeVar, Union, AnyStr, Mapping, Iterable, Optional, AsyncGenerator

import aiohttp

from . import serialization
from .compat import *

if pd:
    PointType = TypeVar('PointType', Mapping, dict, bytes, pd.DataFrame)
    ResultType = TypeVar('ResultType', dict, bytes, pd.DataFrame)
else:
    PointType = TypeVar('PointType', Mapping, dict, bytes)
    ResultType = TypeVar('ResultType', dict, bytes)

# Aioinflux uses logging mainly for debugging purposes.
# Please attach your own handlers if you need logging.
logger = logging.getLogger('aioinflux')


def runner(coro):
    """Function execution decorator."""

    @wraps(coro)
    def inner(self, *args, **kwargs):
        if self.mode == 'async':
            return coro(self, *args, **kwargs)
        return self._loop.run_until_complete(coro(self, *args, **kwargs))

    return inner


class InfluxDBError(Exception):
    """Raised when an server-side error occurs"""
    pass


class InfluxDBWriteError(InfluxDBError):
    """Raised when a server-side writing error occurs"""

    def __init__(self, resp):
        self.status = resp.status
        self.headers = resp.headers
        self.reason = resp.reason
        super().__init__(f'Error writing data ({self.status} - {self.reason}): '
                         f'{self.headers.get("X-Influxdb-Error", "")}')


class InfluxDBClient:
    def __init__(
        self,
        host: str = 'localhost',
        port: int = 8086,
        path: str = '/',
        mode: str = 'async',
        output: str = 'json',
        db: Optional[str] = None,
        database: Optional[str] = None,
        ssl: bool = False,
        *,
        unix_socket: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        timeout: Optional[Union[aiohttp.ClientTimeout, float]] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        redis_opts: Optional[dict] = None,
        cache_expiry: int = 86400,
        **kwargs
    ):
        """
        :class:`~aioinflux.client.InfluxDBClient`  holds information necessary
        to interact with InfluxDB.
        It is async by default, but can also be used as a sync/blocking client.
        When querying, responses are returned as parsed JSON by default,
        but can also be wrapped in easily iterable
        wrapper object or be parsed to Pandas DataFrames.
        The three main public methods are the three endpoints of the InfluxDB API, namely:

        1. :meth:`~.InfluxDBClient.ping`
        2. :meth:`~.InfluxDBClient.write`
        3. :meth:`~.InfluxDBClient.query`

        See each of the above methods documentation for further usage details.

        See also: https://docs.influxdata.com/influxdb/latest/tools/api/

        :param host: Hostname to connect to InfluxDB.
        :param port: Port to connect to InfluxDB.
        :param path: Path to connect to InfluxDB.
        :param mode: Mode in which client should run. Available options:

           - ``async``: Default mode. Each query/request to the backend will
           - ``blocking``: Behaves in sync/blocking fashion,
             similar to the official InfluxDB-Python client.

        :param output: Output format of the response received from InfluxDB.

           - ``json``: Default format.
             Returns parsed JSON as received from InfluxDB.
           - ``dataframe``: Parses results into :py:class`pandas.DataFrame`.
             Not compatible with chunked responses.

        :param db: Default database to be used by the client.
        :param ssl: If https should be used.
        :param unix_socket: Path to the InfluxDB Unix domain socket.
        :param username: Username to use to connect to InfluxDB.
        :param password: User password.
        :param timeout: Timeout in seconds or :class:`aiohttp.ClientTimeout` object
        :param database: Default database to be used by the client.
            This field is for argument consistency with the official InfluxDB Python client.
        :param loop: Asyncio event loop.
        :param redis_opts: Dict fo keyword arguments for :func:`aioredis.create_redis`
        :param cache_expiry: Expiry time (in seconds) for cached data
        :param kwargs: Additional kwargs for :class:`aiohttp.ClientSession`
        """
        self._loop = loop or asyncio.get_event_loop()
        self._session: aiohttp.ClientSession = None
        self._redis: aioredis.Redis = None
        self._mode = None
        self._output = None
        self._db = None
        self.ssl = ssl
        self.host = host
        self.port = port
        self.path = path
        self.mode = mode
        self.output = output
        self.db = database or db

        # ClientSession configuration
        if username:
            kwargs.update(auth=aiohttp.BasicAuth(username, password))
        if unix_socket:
            kwargs.update(connector=aiohttp.UnixConnector(unix_socket, loop=self._loop))
        if timeout:
            if isinstance(timeout, aiohttp.ClientTimeout):
                kwargs.update(timeout=timeout)
            else:
                kwargs.update(timeout=aiohttp.ClientTimeout(total=timeout))
        self.opts = kwargs

        # Cache configuration
        self.redis_opts = redis_opts
        self.cache_expiry = cache_expiry

    async def create_session(self, **kwargs):
        """Creates an :class:`aiohttp.ClientSession`

        Override this or call it with ``kwargs`` to use other :mod:`aiohttp`
        functionality not covered by :class:`~.InfluxDBClient.__init__`
        """
        self.opts.update(kwargs)
        self._session = aiohttp.ClientSession(**self.opts, loop=self._loop)
        if self.redis_opts:
            if aioredis:
                self._redis = await aioredis.create_redis(**self.redis_opts,
                                                          loop=self._loop)
            else:
                warnings.warn(no_redis_warning)

    @property
    def url(self):
        protocol = "https" if self.ssl else "http"
        return f"{protocol}://{self.host}:{self.port}{self.path}{{endpoint}}"

    @property
    def mode(self):
        return self._mode

    @property
    def output(self):
        return self._output

    @property
    def db(self):
        return self._db

    @mode.setter
    def mode(self, mode):
        if mode not in ('async', 'blocking'):
            raise ValueError('Invalid running mode')
        self._mode = mode

    @output.setter
    def output(self, output):
        if pd is None and output == 'dataframe':
            raise ValueError(no_pandas_warning)
        if output not in ('json', 'dataframe'):
            raise ValueError('Invalid output format')
        self._output = output

    @db.setter
    def db(self, db):
        self._db = db
        if not db:
            warnings.warn(f'No default databases is set. '
                          f'Database must be specified when querying/writing.')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    def __del__(self):
        if not self._loop.is_closed() and self._session:
            asyncio.ensure_future(self._session.close(), loop=self._loop)

    def __repr__(self):
        items = [f'{k}={v}' for k, v in vars(self).items() if not k.startswith('_')]
        items.append(f'mode={self.mode}')
        return f'{type(self).__name__}({", ".join(items)})'

    @runner
    async def close(self):
        if self._session:
            await self._session.close()
            self._session = None
        if self._redis:
            self._redis.close()

    @runner
    async def ping(self) -> dict:
        """Pings InfluxDB

         Returns a dictionary containing the headers of the response from ``influxd``.
         """
        if not self._session:
            await self.create_session()
        async with self._session.get(self.url.format(endpoint='ping')) as resp:
            logger.debug(f'{resp.status}: {resp.reason}')
            return dict(resp.headers.items())

    @runner
    async def write(
        self,
        data: Union[PointType, Iterable[PointType]],
        measurement: Optional[str] = None,
        db: Optional[str] = None,
        precision: Optional[str] = None,
        rp: Optional[str] = None,
        tag_columns: Optional[Iterable] = None,
        **extra_tags,
    ) -> bool:
        """Writes data to InfluxDB.
        Input can be:

        1. A mapping (e.g. ``dict``) containing the keys:
           ``measurement``, ``time``, ``tags``, ``fields``
        2. A Pandas :class:`~pandas.DataFrame` with a :class:`~pandas.DatetimeIndex`
        3. A user defined class decorated w/
            :func:`~aioinflux.serialization.usertype.lineprotocol`
        4. A string (``str`` or ``bytes``) properly formatted in InfluxDB's line protocol
        5. An iterable of one of the above

        Input data in formats 1-3 are parsed to the line protocol before being
        written to InfluxDB.
        See the `InfluxDB docs <https://docs.influxdata.com/influxdb/latest/
        write_protocols/line_protocol_reference/>`_ for more details.

        :param data: Input data (see description above).
        :param measurement: Measurement name. Mandatory when when writing DataFrames only.
            When writing dictionary-like data, this field is treated as the default value
            for points that do not contain a `measurement` field.
        :param db: Database to be written to. Defaults to `self.db`.
        :param precision: Sets the precision for the supplied Unix time values.
            Ignored if input timestamp data is of non-integer type.
            Valid values: ``{'ns', 'u', 'µ', 'ms', 's', 'm', 'h'}``
        :param rp: Sets the target retention policy for the write.
            If unspecified, data is written to the default retention policy.
        :param tag_columns: Columns to be treated as tags
            (used when writing DataFrames only)
        :param extra_tags: Additional tags to be added to all points passed.
        :return: Returns ``True`` if insert is successful.
            Raises :py:class:`ValueError` otherwise.
        """
        if not self._session:
            await self.create_session()
        if precision is not None:
            # FIXME: Implement. Related issue: aioinflux/pull/13
            raise NotImplementedError("'precision' parameter is not supported yet")
        data = serialization.serialize(data, measurement, tag_columns, **extra_tags)
        params = {'db': db or self.db}
        if rp:
            params['rp'] = rp
        url = self.url.format(endpoint='write')
        async with self._session.post(url, params=params, data=data) as resp:
            if resp.status == 204:
                return True
            raise InfluxDBWriteError(resp)

    @runner
    async def query(
        self,
        q: AnyStr,
        *,
        epoch: str = 'ns',
        chunked: bool = False,
        chunk_size: Optional[int] = None,
        db: Optional[str] = None,
        use_cache: bool = False,
    ) -> Union[AsyncGenerator[ResultType, None], ResultType]:
        """Sends a query to InfluxDB.
        Please refer to the InfluxDB documentation for all the possible queries:
        https://docs.influxdata.com/influxdb/latest/query_language/

        :param q: Raw query string
        :param db: Database to be queried. Defaults to `self.db`.
        :param epoch: Precision level of response timestamps.
            Valid values: ``{'ns', 'u', 'µ', 'ms', 's', 'm', 'h'}``.
        :param chunked: If ``True``, makes InfluxDB return results in streamed batches
            rather than as a single response.
            Returns an AsyncGenerator which yields responses
            in the same format as non-chunked queries.
        :param chunk_size: Max number of points for each chunk. By default, InfluxDB chunks
            responses by series or by every 10,000 points, whichever occurs first.
        :param use_cache:
        :return: Response in the format specified by the combination of
           :attr:`.InfluxDBClient.output` and ``chunked``
        """

        async def _chunked_generator(url, data, dataframe):
            async with self._session.post(url, data=data) as resp:
                logger.debug(f'{resp.status} (CHUNKED): {q!r}')
                # Hack to avoid aiohttp raising ValueError('Line is too long')
                # The number 128 is arbitrary (may be too large/small).
                resp.content._high_water *= 128
                chunk_count = 0
                async for chunk in resp.content:
                    chunk = json.loads(chunk)
                    self._check_error(chunk)
                    chunk_count += 1
                    logger.debug(f'Yielding chunk #{chunk_count:03d}')
                    if dataframe:
                        yield serialization.dataframe.parse(chunk)
                    else:
                        yield chunk

        if not self._session:
            await self.create_session()

        # InfluxDB documentation is wrong regarding `/query` parameters
        # See https://github.com/influxdata/docs.influxdata.com/issues/1807
        if not isinstance(chunked, bool):
            raise ValueError("'chunked' must be a boolean")
        data = dict(q=q, db=db or self.db, chunked=str(chunked).lower(), epoch=epoch)
        if chunked and chunk_size:
            data['chunk_size'] = chunk_size

        url = self.url.format(endpoint='query')
        if chunked:
            if use_cache:
                raise ValueError("Can't use cache w/ chunked queries")
            elif self.mode != 'async':
                raise ValueError("Can't use 'chunked' with non-async mode")
            else:
                return _chunked_generator(url, data, self.output == 'dataframe')

        key = f'aioinflux:{q}'
        if use_cache and self._redis and await self._redis.exists(key):
            logger.debug(f'Cache HIT: {q}')
            data = lz4.decompress(await self._redis.get(key))
        else:
            async with self._session.post(url, data=data) as resp:
                data = await resp.read()
                if use_cache and self._redis:
                    logger.debug(f'Cache MISS ({resp.status}): {q}')
                    if resp.status == 200:
                        await self._redis.set(key, lz4.compress(data))
                        await self._redis.expire(key, self.cache_expiry)
                else:
                    logger.debug(f'{resp.status}: {q}')

        data = json.loads(data)
        self._check_error(data)
        if self.output == 'json':
            return data
        elif self.output == 'dataframe':
            return serialization.dataframe.parse(data)
        else:
            raise ValueError('Invalid output format')

    @staticmethod
    def _check_error(response):
        """Checks for JSON error messages and raises Python exception"""
        if 'error' in response:
            raise InfluxDBError(response['error'])
        elif 'results' in response:
            for statement in response['results']:
                if 'error' in statement:
                    msg = '{d[error]} (statement {d[statement_id]})'
                    raise InfluxDBError(msg.format(d=statement))

    # InfluxQL - Data management
    # --------------------------

    def create_database(self, db=None):
        db = db or self.db
        return self.query(f'CREATE DATABASE "{db}"')

    def drop_database(self, db=None):
        db = db or self.db
        return self.query(f'DROP DATABASE "{db}"')

    def drop_measurement(self, measurement):
        return self.query(f'DROP MEASUREMENT "{measurement}"')

    # InfluxQL - Schema exploration
    # -----------------------------

    def show_databases(self):
        return self.query("SHOW DATABASES")

    def show_measurements(self):
        return self.query("SHOW MEASUREMENTS")

    def show_users(self):
        return self.query("SHOW USERS")

    def show_series(self, measurement=None):
        if measurement:
            return self.query(f"SHOW SERIES FROM {measurement}")
        return self.query("SHOW SERIES")

    def show_tag_keys(self, measurement=None):
        if measurement:
            return self.query(f"SHOW TAG KEYS FROM {measurement}")
        return self.query("SHOW TAG KEYS")

    def show_field_keys(self, measurement=None):
        if measurement:
            return self.query(f"SHOW FIELD KEYS FROM {measurement}")
        return self.query("SHOW FIELD KEYS")

    def show_tag_values(self, key, measurement=None):
        if measurement:
            return self.query(f'SHOW TAG VALUES FROM "{measurement}" WITH key = "{key}"')
        return self.query(f'SHOW TAG VALUES WITH key = "{key}"')

    def show_retention_policies(self):
        return self.query("SHOW RETENTION POLICIES")

    # InfluxQL - Other
    # ----------------

    def show_continuous_queries(self):
        return self.query("SHOW CONTINUOUS QUERIES")



import asyncio

import pytest

from aioinflux import InfluxDBClient
import testing_utils as utils


@pytest.yield_fixture(scope='module')
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope='module')
async def client():
    async with InfluxDBClient(db='client_test', mode='async') as client:
        await client.create_database()
        yield client
        await client.drop_database()


@pytest.fixture(scope='module')
async def cache_client():
    opts = dict(
        db='cache_client_test',
        redis_opts=dict(
            address='redis://localhost:6379/8',
            timeout=5,
        ),
        cache_expiry=600
    )
    async with InfluxDBClient(**opts, mode='async') as client:
        assert await client.create_database()
        yield client
        await client.drop_database()
        await client._redis.flushdb()


@pytest.fixture(scope='module')
def df_client():
    if utils.pd is None:
        return
    with InfluxDBClient(db='df_client_test', mode='blocking', output='dataframe') as client:
        client.create_database()
        yield client
        client.drop_database()


@pytest.fixture(scope='module')
async def iter_client():
    async with InfluxDBClient(db='iter_client_test', mode='async') as client:
        await client.create_database()
        await client.write([p for p in utils.cpu_load_generator(100)])
        yield client
        await client.drop_database()



# flake8: noqa
import uuid
import enum
from datetime import datetime
from typing import NamedTuple, Optional
from collections import namedtuple
from dataclasses import dataclass
from decimal import Decimal

import pytest

import aioinflux
from aioinflux import lineprotocol, SchemaError
from testing_utils import logger


class CpuLoad(enum.Enum):
    LOW = 10
    HIGH = 100


@pytest.mark.asyncio
async def test_decorator(client):
    @lineprotocol
    class MyPoint(NamedTuple):
        measurement: aioinflux.MEASUREMENT
        time: aioinflux.TIMEINT
        host: aioinflux.TAG
        running: aioinflux.BOOL
        users: aioinflux.INT
        cpu_load: aioinflux.FLOAT
        cpu_load_level: aioinflux.ENUM
        cpu_load_level_tag: aioinflux.TAGENUM
        running_cost: aioinflux.DECIMAL
        uuid: aioinflux.STR

    p = MyPoint(
        measurement="dp",
        time=1500,
        host="us1",
        running=True,
        users=1000,
        cpu_load=99.5,
        cpu_load_level=CpuLoad.HIGH,
        cpu_load_level_tag=CpuLoad.LOW,
        running_cost=Decimal('3.54'),
        uuid=str(uuid.uuid4()),
    )
    assert p
    assert hasattr(p, 'to_lineprotocol')
    assert await client.write(p)
    logger.info(await client.query('SELECT * FROM dp'))
    logger.info(await client.query("SHOW FIELD KEYS FROM dp"))


def test_functional():
    schema = dict(
        measurement=aioinflux.MEASUREMENT,
        time=aioinflux.TIMEINT,
        host=aioinflux.TAG,
        running=aioinflux.BOOL,
        users=aioinflux.INT,
    )
    MyPoint = lineprotocol(namedtuple('MyPoint', schema.keys()), schema=schema)
    p = MyPoint("a", 2, "b", False, 5)
    logger.debug(p.to_lineprotocol())
    assert isinstance(p.to_lineprotocol(), bytes)


def test_datestr():
    schema = dict(
        measurement=aioinflux.MEASUREMENT,
        time=aioinflux.TIMESTR,
        host=aioinflux.TAG,
        running=aioinflux.BOOL,
        users=aioinflux.INT,
    )
    MyPoint = lineprotocol(namedtuple('MyPoint', schema.keys()), schema=schema)
    p = MyPoint("a", "2018-08-08 15:22:33", "b", False, 5)
    logger.debug(p.to_lineprotocol())
    assert isinstance(p.to_lineprotocol(), bytes)


def test_datetime():
    schema = dict(
        measurement=aioinflux.MEASUREMENT,
        time=aioinflux.TIMEDT,
        host=aioinflux.TAG,
        running=aioinflux.BOOL,
        users=aioinflux.INT,
    )
    MyPoint = lineprotocol(namedtuple('MyPoint', schema.keys()), schema=schema)
    p = MyPoint("a", datetime.utcnow(), "b", False, 5)
    logger.debug(p.to_lineprotocol())
    assert isinstance(p.to_lineprotocol(), bytes)


def test_placeholder():
    @lineprotocol(placeholder=True)
    @dataclass
    class MyPoint:
        timestamp: aioinflux.TIMEINT

    lp = MyPoint(0).to_lineprotocol()
    logger.debug(lp)


def test_extra_tags():
    @lineprotocol(extra_tags={'host': 'ap1'})
    class MyPoint(NamedTuple):
        measurement: aioinflux.MEASUREMENT
        time: aioinflux.TIMEINT
        running: aioinflux.BOOL
        users: aioinflux.INT

    p = MyPoint("a", 2, False, 5)
    assert b'ap1' in p.to_lineprotocol()


def test_rm_none():
    @lineprotocol(rm_none=True)
    class MyPoint(NamedTuple):
        measurement: aioinflux.MEASUREMENT
        time: aioinflux.TIMEINT
        host: aioinflux.TAG
        running: Optional[aioinflux.BOOL]
        users: Optional[aioinflux.INT]

    p1= MyPoint("a", 2, "b", True, None)
    p2 = MyPoint("a", 2, "b", None, 1)
    logger.debug(p1.to_lineprotocol())
    logger.debug(p2.to_lineprotocol())
    assert b'users' not in p1.to_lineprotocol()
    assert b'running' not in p2.to_lineprotocol()


# noinspection PyUnusedLocal
def test_schema_error():
    with pytest.raises(SchemaError):
        @lineprotocol
        class MyPoint:
            pass

    with pytest.raises(SchemaError):
        @lineprotocol  # noqa: F811
        class MyPoint(NamedTuple):
            measurement: aioinflux.MEASUREMENT
            time: aioinflux.TIMEINT
            host: aioinflux.TAG
            running: bool
            users: aioinflux.INT

    with pytest.raises(SchemaError):
        @lineprotocol  # noqa: F811
        class MyPoint(NamedTuple):
            measurement: aioinflux.MEASUREMENT
            measurement2: aioinflux.MEASUREMENT
            time: aioinflux.TIMEINT
            host: aioinflux.TAG
            running: aioinflux.BOOL
            users: aioinflux.INT

    with pytest.raises(SchemaError):
        @lineprotocol  # noqa: F811
        class MyPoint(NamedTuple):
            measurement: aioinflux.MEASUREMENT
            time: aioinflux.TIMEINT
            time2: aioinflux.TIMEDT
            host: aioinflux.TAG
            running: aioinflux.BOOL
            users: aioinflux.INT

    with pytest.raises(SchemaError):
        @lineprotocol  # noqa: F811
        class MyPoint(NamedTuple):
            measurement: aioinflux.MEASUREMENT
            time: aioinflux.TIMEINT
            host: aioinflux.TAG



import warnings

# Special characters documentation:
# https://docs.influxdata.com/influxdb/v1.4/write_protocols/line_protocol_reference/#special-characters
# Although not in the official docs, new line characters are removed in order to avoid issues.
# Go implementation: https://github.com/influxdata/influxdb/blob/master/pkg/escape/strings.go
key_escape = str.maketrans({'\\': '\\\\', ',': r'\,', ' ': r'\ ', '=': r'\=', '\n': ''})
tag_escape = str.maketrans({'\\': '\\\\', ',': r'\,', ' ': r'\ ', '=': r'\=', '\n': ''})
str_escape = str.maketrans({'\\': '\\\\', '"': r'\"', '\n': ''})
measurement_escape = str.maketrans({'\\': '\\\\', ',': r'\,', ' ': r'\ ', '\n': ''})


def escape(string, escape_pattern):
    """Assistant function for string escaping"""
    try:
        return string.translate(escape_pattern)
    except AttributeError:
        warnings.warn("Non-string-like data passed. "
                      "Attempting to convert to 'str'.")
        return str(string).translate(tag_escape)



import pytest
from testing_utils import logger, requires_redis


@requires_redis
@pytest.mark.asyncio
async def test_cache(cache_client):
    assert await cache_client.write('foo bar=1')
    q = 'SELECT bar from foo'

    r1 = await cache_client.query(q, use_cache=True)  # Create cache
    logger.debug(r1)
    assert await cache_client._redis.exists(f'aioinflux:{q}')

    r2 = await cache_client.query(q, use_cache=True)  # Get cache
    logger.debug(r2)
    r3 = await cache_client.query(q, use_cache=False)  # Ignore cache

    r1 = r1['results'][0]['series'][0]
    r2 = r2['results'][0]['series'][0]
    r3 = r3['results'][0]['series'][0]
    assert r1 == r2 == r3



import pytest
from aioinflux import iterpoints
from testing_utils import logger


@pytest.mark.asyncio
async def test_iterpoints_with_parser(iter_client):
    r = await iter_client.query("SELECT * FROM cpu_load LIMIT 3")
    parser = lambda *x, meta: dict(zip(meta['columns'], x))  # noqa
    for i in iterpoints(r, parser):
        logger.info(i)
        assert 'time' in i
        assert 'value' in i
        assert 'host' in i


@pytest.mark.asyncio
async def test_aiter_point(iter_client):
    resp = await iter_client.query('SELECT * from cpu_load', chunked=True, chunk_size=10)
    points = []
    async for chunk in resp:
        for point in iterpoints(chunk):
            points.append(point)
    assert len(points) == 100


@pytest.mark.asyncio
async def test_iter_point_namedtuple(iter_client):
    from collections import namedtuple
    nt = namedtuple('cpu_load', ['time', 'direction', 'host', 'region', 'value'])

    resp = await iter_client.query('SELECT * from cpu_load')
    points = []
    for point in iterpoints(resp, parser=nt):
        points.append(point)
        assert len(point) == 5
    assert len(points) == 100



# flake8: noqa
from . import serialization
from .client import InfluxDBClient, InfluxDBError, InfluxDBWriteError
from .iterutils import iterpoints
from .serialization.usertype import *

__version__ = '0.9.0'



from datetime import datetime

import pytz
import pytest
from aioinflux.serialization.mapping import _serialize_timestamp
from testing_utils import logger
from aioinflux.compat import pd


def test_timestamp_timezone_parsing():
    dt_naive = datetime(2018, 1, 1)
    dt_aware = datetime(2018, 1, 1, tzinfo=pytz.UTC)
    str_naive = str(dt_naive)
    str_aware = str(dt_aware)

    for i in [dt_naive, dt_aware, str_naive, str_aware]:
        assert _serialize_timestamp({'time': i}) == 1514764800000000000


@pytest.mark.skipif(pd is not None, reason='ciso8601-specific test')
def test_invalid_timestamp_parsing():
    with pytest.raises(ValueError) as e:
        _serialize_timestamp({'time': '2018/01/01'})
    logger.error(e)


def test_invalid_timestamp_parsing2():
    with pytest.raises(ValueError) as e:
        _serialize_timestamp({'time': 'foo'})
    logger.error(e)



import warnings

no_pandas_warning = "Pandas/Numpy is not available. Support for 'dataframe' mode is disabled."
no_redis_warning = "Redis dependencies not available. Support for caching is disabled."

try:
    import pandas as pd
    import numpy as np
except ModuleNotFoundError:
    pd = None
    np = None
    warnings.warn(no_pandas_warning)

try:
    import aioredis
    import lz4.block as lz4
except ModuleNotFoundError:
    aioredis = None
    lz4 = None

__all__ = ['no_pandas_warning', 'no_redis_warning', 'pd', 'np', 'aioredis', 'lz4']



import time
from typing import Mapping

import ciso8601

from .common import *


def serialize(point: Mapping, measurement=None, **extra_tags) -> bytes:
    """Converts dictionary-like data into a single line protocol line (point)"""
    tags = _serialize_tags(point, extra_tags)
    return (
        f'{_serialize_measurement(point, measurement)}'
        f'{"," if tags else ""}{tags} '
        f'{_serialize_fields(point)} '
        f'{_serialize_timestamp(point)}'
    ).encode()


def _serialize_measurement(point, measurement):
    try:
        return escape(point['measurement'], measurement_escape)
    except KeyError:
        if measurement is None:
            raise ValueError("'measurement' missing")
        return escape(measurement, measurement_escape)


def _serialize_tags(point, extra_tags):
    output = []
    for k, v in {**point.get('tags', {}), **extra_tags}.items():
        k = escape(k, key_escape)
        v = escape(v, tag_escape)
        if not v:
            continue  # ignore blank/null string tags
        output.append(f'{k}={v}')
    return ','.join(output)


def _serialize_timestamp(point):
    dt = point.get('time')
    if not dt:
        return ''
    elif isinstance(dt, int):
        return dt
    elif isinstance(dt, (str, bytes)):
        dt = ciso8601.parse_datetime(dt)
        if not dt:
            raise ValueError(f'Invalid datetime string: {dt!r}')

    if not dt.tzinfo:
        # Assume tz-naive input to be in UTC, not local time
        return int(dt.timestamp() - time.timezone) * 10 ** 9 + dt.microsecond * 1000
    return int(dt.timestamp()) * 10 ** 9 + dt.microsecond * 1000


def _serialize_fields(point):
    """Field values can be floats, integers, strings, or Booleans."""
    output = []
    for k, v in point['fields'].items():
        k = escape(k, key_escape)
        if isinstance(v, bool):
            output.append(f'{k}={v}')
        elif isinstance(v, int):
            output.append(f'{k}={v}i')
        elif isinstance(v, str):
            output.append(f'{k}="{v.translate(str_escape)}"')
        elif v is None:
            # Empty values
            continue
        else:
            # Floats
            output.append(f'{k}={v}')
    return ','.join(output)



import inspect

from typing import Optional, Iterator, Callable, Any


def iterpoints(resp: dict, parser: Optional[Callable] = None) -> Iterator[Any]:
    """Iterates a response JSON yielding data point by point.

    Can be used with both regular and chunked responses.
    By default, returns just a plain list of values representing each point,
    without column names, or other metadata.

    In case a specific format is needed, an optional ``parser`` argument can be passed.
    ``parser`` is a function/callable that takes data point values
    and, optionally, a ``meta`` parameter containing which takes a
    dictionary containing all or a subset of the following:
    ``{'columns', 'name', 'tags', 'statement_id'}``.

    Sample parser functions:

    .. code:: python

       # Function optional meta argument
       def parser(*x, meta):
           return dict(zip(meta['columns'], x))

       # Namedtuple (callable)
       from collections import namedtuple
       parser = namedtuple('MyPoint', ['col1', 'col2', 'col3'])


    :param resp: Dictionary containing parsed JSON (output from InfluxDBClient.query)
    :param parser: Optional parser function/callable
    :return: Generator object
    """
    for statement in resp['results']:
        if 'series' not in statement:
            continue
        for series in statement['series']:
            if parser is None:
                return (x for x in series['values'])
            elif 'meta' in inspect.signature(parser).parameters:
                meta = {k: series[k] for k in series if k != 'values'}
                meta['statement_id'] = statement['statement_id']
                return (parser(*x, meta=meta) for x in series['values'])
            else:
                return (parser(*x) for x in series['values'])
    return iter([])



import enum
import ciso8601
import time
import decimal
from collections import Counter
from typing import TypeVar, Optional, Mapping, Union
from datetime import datetime

# noinspection PyUnresolvedReferences
from .common import *  # noqa
from ..compat import pd

__all__ = [
    'lineprotocol', 'SchemaError',
    'MEASUREMENT', 'TIMEINT', 'TIMESTR', 'TIMEDT',
    'TAG', 'TAGENUM',
    'BOOL', 'INT', 'DECIMAL', 'FLOAT', 'STR', 'ENUM',
]

MEASUREMENT = TypeVar('MEASUREMENT', bound=str)
TIMEINT = TypeVar('TIMEINT', bound=int)
TIMESTR = TypeVar('TIMESTR', bound=str)
TIMEDT = TypeVar('TIMEDT', bound=datetime)
TAG = TypeVar('TAG', bound=str)
TAGENUM = TypeVar('TAGENUM', bound=enum.Enum)
BOOL = TypeVar('BOOL', bound=bool)
INT = TypeVar('INT', bound=int)
DECIMAL = TypeVar('DECIMAL', bound=decimal.Decimal)
FLOAT = TypeVar('FLOAT', bound=float)
STR = TypeVar('STR', bound=str)
ENUM = TypeVar('ENUM', bound=enum.Enum)

time_types = [TIMEINT, TIMEDT, TIMESTR]
tag_types = [TAG, TAGENUM]
field_types = [BOOL, INT, DECIMAL, FLOAT, STR, ENUM]
optional_field_types = [Optional[f] for f in field_types]


class SchemaError(TypeError):
    """Raised when invalid schema is passed to :func:`lineprotocol`"""


def str_to_dt(s):
    dt = ciso8601.parse_datetime(s)
    if dt:
        return dt
    raise ValueError(f'Invalid datetime string: {dt!r}')


def dt_to_int(dt):
    if not dt.tzinfo:
        # Assume tz-naive input to be in UTC, not local time
        return int(dt.timestamp() - time.timezone) * 10 ** 9 + dt.microsecond * 1000
    return int(dt.timestamp()) * 10 ** 9 + dt.microsecond * 1000


def _validate_schema(schema, placeholder):
    c = Counter(schema.values())
    if not c:
        raise SchemaError("Schema/type annotations missing")
    if c[MEASUREMENT] > 1:
        raise SchemaError("Class can't have more than one 'MEASUREMENT' attribute")
    if sum(c[e] for e in time_types) > 1:
        raise SchemaError(f"Can't have more than one timestamp-type attribute {time_types}")
    if sum(c[e] for e in field_types + optional_field_types) < 1 and not placeholder:
        raise SchemaError(f"Must have one or more non-empty "
                          f"field-type attributes {field_types}")


def is_optional(t, base_type):
    """Checks if type hint is Optional[base_type]"""
    # NOTE: The 'typing' module is still "provisional" and documentation sub-optimal,
    #  which requires these kinds instrospection into undocumented implementation details
    # NOTE: May break in Python 3.8
    # TODO: Check if works on Python 3.6
    try:
        cond1 = getattr(t, '__origin__') is Union
        cond2 = {type(None), base_type} == set(getattr(t, '__args__', []))
        if cond1 and cond2:
            return True
    except AttributeError:
        return False
    return False


def _make_serializer(meas, schema, extra_tags, placeholder):  # noqa: C901
    """Factory of line protocol parsers"""
    _validate_schema(schema, placeholder)
    tags = []
    fields = []
    ts = None
    meas = meas
    for k, t in schema.items():
        if t is MEASUREMENT:
            meas = f"{{i.{k}}}"
        elif t is TIMEINT:
            ts = f"{{i.{k}}}"
        elif t is TIMESTR:
            if pd:
                ts = f"{{pd.Timestamp(i.{k} or 0).value}}"
            else:
                ts = f"{{dt_to_int(str_to_dt(i.{k}))}}"
        elif t is TIMEDT:
            if pd:
                ts = f"{{pd.Timestamp(i.{k} or 0).value}}"
            else:
                ts = f"{{dt_to_int(i.{k})}}"
        elif t is TAG or is_optional(t, TAG):
            tags.append(f"{k}={{str(i.{k}).translate(tag_escape)}}")
        elif t is TAGENUM or is_optional(t, TAGENUM):
            tags.append(f"{k}={{getattr(i.{k}, 'name', i.{k} or None)}}")
        elif t is FLOAT or is_optional(t, FLOAT):
            fields.append(f"{k}={{i.{k}}}")
        elif t is DECIMAL or is_optional(t, DECIMAL):
            fields.append(f"{k}={{i.{k}}}")
        elif t is BOOL or is_optional(t, BOOL):
            fields.append(f"{k}={{i.{k}}}")
        elif t is INT or is_optional(t, INT):
            fields.append(f"{k}={{i.{k}}}i")
        elif t is STR or is_optional(t, STR):
            fields.append(f"{k}=\\\"{{str(i.{k}).translate(str_escape)}}\\\"")
        elif t is ENUM or is_optional(t, ENUM):
            fields.append(f"{k}=\\\"{{getattr(i.{k}, 'name', i.{k} or None)}}\\\"")
        else:
            raise SchemaError(f"Invalid attribute type {k!r}: {t!r}")
    extra_tags = extra_tags or {}
    for k, v in extra_tags.items():
        tags.append(f"{k}={v}")
    if placeholder:
        fields.insert(0, f"_=true")

    sep = ',' if tags else ''
    ts = f' {ts}' if ts else ''
    fmt = f"{meas}{sep}{','.join(tags)} {','.join(fields)}{ts}"
    f = eval(f'lambda i: f"{fmt}".encode()')
    f.__doc__ = "Returns InfluxDB line protocol representation of user-defined class"
    return f


def lineprotocol(
        cls=None,
        *,
        schema: Optional[Mapping[str, type]] = None,
        rm_none: bool = False,
        extra_tags: Optional[Mapping[str, str]] = None,
        placeholder: bool = False
):
    """Adds ``to_lineprotocol`` method to arbitrary user-defined classes

    :param cls: Class to monkey-patch
    :param schema: Schema dictionary (attr/type pairs).
    :param rm_none: Whether apply a regex to remove ``None`` values.
        If ``False``, passing ``None`` values to boolean, integer or float or time fields
        will result in write errors. Setting to ``True`` is "safer" but impacts performance.
    :param extra_tags: Hard coded tags to be added to every point generated.
    :param placeholder: If no field attributes are present, add a placeholder attribute (``_``)
        which is always equal to ``True``. This is a workaround for creating field-less points
        (which is not supported natively by InfluxDB)
    """
    opts = dict(
        schema=schema,
        rm_none=rm_none,
        extra_tags=extra_tags or {},
        placeholder=placeholder,
    )

    def _lineprotocol(cls):
        _schema = schema or getattr(cls, '__annotations__', {})
        # TODO: Raise warning or exception if schema has optionals but rm_none is False
        # for t in _schema.values():
        #     for bt in field_types + tag_types:
        #         if is_optional(t, bt):
        #             warnings.warn("")
        f = _make_serializer(cls.__name__, _schema, extra_tags, placeholder)
        cls.to_lineprotocol = f
        cls.to_lineprotocol.opts = opts
        return cls

    def _rm_none_lineprotocol(cls):

        def _parser_selector(i):
            if not hasattr(i, '_asdict'):
                raise ValueError("'rm_none' can only be used with namedtuples")
            key = tuple([k for k, v in i._asdict().items() if v != '' and v is not None])
            if key not in parsers:
                _schema = schema or getattr(cls, '__annotations__', {})
                _schema = {k: v for k, v in _schema.items() if k in key}
                parsers[key] = _make_serializer(cls.__name__, _schema, extra_tags, placeholder)
            return parsers[key](i)

        parsers = {}
        cls.to_lineprotocol = _parser_selector
        cls.to_lineprotocol.opts = opts
        return cls

    if cls:
        if rm_none:
            # Using rm_none has substantial runtime impact.
            # Best avoided if performance is critical.
            return _rm_none_lineprotocol(cls)
        # No options
        return _lineprotocol(cls)
    else:
        if rm_none:
            return _rm_none_lineprotocol
        return _lineprotocol



import pytest
from aioinflux import InfluxDBClient, InfluxDBError, InfluxDBWriteError, iterpoints
from aioinflux.compat import pd
import testing_utils as utils
from testing_utils import logger


def test_repr(client):
    logger.info(client)


@pytest.mark.asyncio
async def test_ping(client):
    r = await client.ping()
    assert 'X-Influxdb-Version' in r


#########
# Write #
#########

@pytest.mark.asyncio
async def test_write_simple(client):
    assert await client.write(utils.random_points(100))


@pytest.mark.asyncio
async def test_write_string(client):
    point = 'cpu_load_short,host=server02,region=us-west value=0.55 1422568543702900257'
    assert await client.write(point)


@pytest.mark.asyncio
async def test_write_tagless(client):
    point = b'cpu_load_short value=0.55 1423568543000000000'
    assert await client.write(point)


@pytest.mark.asyncio
async def test_write_special_values(client):
    point = utils.random_point()
    point['tags']['boolean_tag'] = True
    point['tags']['none_tag'] = None
    point['tags']['blank_tag'] = ''
    point['fields']['boolean_field'] = False
    point['fields']['none_field'] = None
    point['fields']['backslash'] = "This is a backslash: \\"
    point['measurement'] = '"quo⚡️es and emoji"'
    with pytest.warns(UserWarning) as e:
        assert await client.write(point)
    logger.warning(e)


@pytest.mark.asyncio
async def test_write_with_custom_measurement(client):
    points = [p for p in utils.random_points(5)]
    for p in points:
        _ = p.pop('measurement')
    logger.info(points)
    with pytest.raises(ValueError):
        assert await client.write(points)
    assert await client.write(points, measurement='another_measurement')
    resp = await client.query('SELECT * FROM another_measurement')
    assert len(resp['results'][0]['series'][0]['values']) == 5


@pytest.mark.asyncio
async def test_write_without_timestamp(client):
    points = [p for p in utils.random_points(9)]
    for p in points:
        _ = p.pop('time')
        _ = p.pop('measurement')
    logger.info(points)
    assert await client.write(points, measurement='yet_another_measurement')
    resp = await client.query('SELECT * FROM yet_another_measurement')
    # Points with the same tag/timestamp set are overwritten
    assert len(resp['results'][0]['series'][0]['values']) == 1


@pytest.mark.asyncio
async def test_write_non_string_identifier_and_tags(client):
    point = dict(tags={1: 2},
                 fields={3: 4})
    with pytest.warns(UserWarning):
        assert await client.write(point, measurement='my_measurement')
    resp = await client.query('SELECT * FROM my_measurement')
    logger.info(resp)
    assert len(resp['results'][0]['series'][0]['values']) == 1


@pytest.mark.asyncio
async def test_write_to_non_default_db(client):
    points = [p for p in utils.random_points(5)]
    await client.create_database(db='temp_db')
    assert client.db != 'temp_db'
    assert await client.write(points, db='temp_db')
    resp = await client.query('SELECT * FROM temp_db..test_measurement')
    logger.info(resp)
    assert len(resp['results'][0]['series'][0]['values']) == 5
    await client.drop_database(db='temp_db')


@pytest.mark.asyncio
async def test_write_to_non_default_rp(client):
    db = client.db
    await client.query(f"CREATE RETENTION POLICY myrp ON {db} DURATION 1h REPLICATION 1")
    points = [p for p in utils.random_points(5)]
    assert await client.write(points, rp='myrp')
    resp = await client.query(f"SELECT * from {db}.myrp.test_measurement")
    logger.info(resp)
    assert len(resp['results'][0]['series'][0]['values']) == 5


#########
# Query #
#########

@pytest.mark.asyncio
async def test_simple_query(client):
    resp = await client.query('SELECT * FROM test_measurement')
    assert len(resp['results'][0]['series'][0]['values']) == 100


@pytest.mark.asyncio
async def test_chunked_query(client):
    resp = await client.query('SELECT * FROM test_measurement',
                              chunked=True, chunk_size=10)
    points = []
    async for chunk in resp:
        for point in iterpoints(chunk):
            points.append(point)
    assert len(points) == 100


@pytest.mark.asyncio
async def test_empty_chunked_query(client):
    resp = await client.query('SELECT * FROM fake', chunked=True, chunk_size=10)
    points = []
    async for chunk in resp:
        for point in iterpoints(chunk):
            points.append(point)
    assert len(points) == 0


####################
# Built-in queries #
####################


@pytest.mark.asyncio
async def test_create_database(client):
    resp = await client.create_database(db='mytestdb')
    assert resp


@pytest.mark.asyncio
async def test_drop_database(client):
    resp = await client.drop_database(db='mytestdb')
    assert resp


@pytest.mark.asyncio
async def test_drop_measurement(client):
    measurement = utils.random_string()
    assert await client.write(f'{measurement} foo=1')
    await client.drop_measurement(measurement=measurement)


@pytest.mark.asyncio
async def test_show_databases(client):
    r = await client.show_databases()
    assert r
    logger.debug(r)


@pytest.mark.asyncio
async def test_show_measurements(client):
    r = await client.show_measurements()
    assert r
    logger.debug(r)


@pytest.mark.asyncio
async def test_show_users(client):
    r = await client.show_users()
    assert r
    logger.debug(r)


@pytest.mark.asyncio
async def test_show_series(client):
    r = await client.show_series()
    assert r
    logger.debug(r)
    r = await client.show_series('cpu_load_short')
    assert r
    logger.debug(r)


@pytest.mark.asyncio
async def test_show_retention_policies(client):
    r = await client.show_retention_policies()
    assert r
    logger.debug(r)


@pytest.mark.asyncio
async def test_show_tag_keys(client):
    r = await client.show_tag_keys()
    assert r
    logger.debug(r)
    r = await client.show_tag_keys('cpu_load_short')
    assert r
    logger.debug(r)


@pytest.mark.asyncio
async def test_show_field_keys(client):
    r = await client.show_field_keys()
    assert r
    logger.debug(r)
    r = await client.show_field_keys('cpu_load_short')
    assert r
    logger.debug(r)


@pytest.mark.asyncio
async def test_show_tag_values(client):
    r = await client.show_tag_values('host')
    assert r
    logger.debug(r)
    r = await client.show_tag_values('host', 'cpu_load_short')
    assert r
    logger.debug(r)


@pytest.mark.asyncio
async def test_show_continuous_queries(client):
    r = await client.show_continuous_queries()
    assert r
    logger.debug(r)


###############
# Error tests #
###############

@pytest.mark.asyncio
async def test_chunked_query_error(client):
    with pytest.raises(InfluxDBError) as e:
        resp = await client.query('INVALID QUERY', chunked=True, chunk_size=10)
        _ = [i async for i in resp]
    logger.error(e)


@pytest.mark.asyncio
async def test_invalid_data_write(client):
    with pytest.raises(InfluxDBWriteError) as e:
        # Plain invalid data
        await client.write(utils.random_string())
    logger.error(e)

    with pytest.raises(ValueError) as e:
        # Pass function as input data
        await client.write(utils.random_string)
    logger.error(e)

    with pytest.raises(ValueError) as e:
        # Measurement missing
        point = utils.random_point()
        point.pop('measurement')
        await client.write(point)
    logger.error(e)


def test_invalid_client_mode():
    with pytest.raises(ValueError) as e:
        _ = InfluxDBClient(db='mytestdb', mode=utils.random_string())
    logger.error(e)


def test_no_default_database_warning():
    with pytest.warns(UserWarning) as e:
        _ = InfluxDBClient(db=None)
    logger.error(e)


def test_invalid_output_format(client):
    with pytest.raises(ValueError) as e:
        client.output = utils.random_string()
    logger.error(e)
    if pd is None:
        with pytest.raises(ValueError) as e:
            client.output = 'dataframe'
        logger.error(e)


@pytest.mark.asyncio
async def test_invalid_query(client):
    with pytest.raises(InfluxDBError) as e:
        await client.query('NOT A VALID QUERY')
    logger.error(e)


@pytest.mark.asyncio
async def test_statement_error(client):
    with pytest.raises(InfluxDBError) as e:
        await client.query('SELECT * FROM my_measurement', db='fake_db')
    logger.error(e)



import datetime
import logging
import random
import string
import uuid
from itertools import combinations, cycle, islice

from aioinflux.compat import *

import pytest

requires_pandas = pytest.mark.skipif(pd is None, reason=no_pandas_warning)
requires_redis = pytest.mark.skipif(aioredis is None, reason=no_redis_warning)
logger = logging.getLogger('aioinflux')


def random_point():
    now = datetime.datetime.now()
    point = {
        'measurement': 'test_measurement',  # noqa
        'tags': {'tag key with sp🚀ces': 'tag,value,with"commas"'},
        'time': random.choice([now, str(now)]),
        'fields': {
            'fi\neld_k\ey': random.randint(0, 200),
            'quote': '"',
            'value': random.random(),
        }
    }
    return point


def random_points(n=10):
    for i in range(n):
        yield random_point()


def random_dataframe():
    """Generates a DataFrame with five random walk columns and a tag column"""
    arr = np.cumsum(np.random.randn(50, 5), axis=1)
    letters = combinations(string.ascii_uppercase, 3)
    columns = [''.join(triplet) for triplet in random.choices(list(letters), k=5)]
    tags = [chr(i + 65) for i in np.random.randint(0, 5, 50)]
    ix = pd.date_range(end=pd.Timestamp.utcnow(), periods=50, freq='90min')

    df = pd.DataFrame(arr, columns=columns)
    df['tag'] = tags
    df['noise'] = list(islice(cycle(["a", '\n', r"\n"]), 50))
    df.index = ix
    return df


def trading_df(n=100):
    sym = [''.join(i) for i in combinations('ABCDE', 3)]

    df = pd.DataFrame({
        'str_id': [str(uuid.uuid4()) for _ in range(n)],
        'px': 1000 + np.cumsum(np.random.randint(-10, 11, n)) / 2,
        'sym': np.random.choice(sym, n),
        'side': np.random.choice(['BUY', 'SELL'], n),
        'size': np.random.randint(1, 10, size=n) * 100,
        'valid': np.random.randint(2, size=n).astype('bool')
    })
    df.index = pd.date_range(end=pd.Timestamp.utcnow(), periods=n, freq='1s')
    df['side'] = df['side'].astype('category')
    return df


def random_string():
    return ''.join(random.choices(string.ascii_lowercase, k=random.randint(4, 10)))


def cpu_load_generator(n):
    p = 'cpu_load,direction={d},host=server{s:02d},region=us-{r} value={f:.5f} {t}'
    t = 1520535379386016000
    d = ['in', 'out']
    r = ['north', 'south', 'west', 'east']
    for _ in range(n):
        t += random.randint(1, 10 ** 10)
        yield p.format(
            t=t,
            d=random.choice(d),
            r=random.choice(r),
            s=random.randint(1, 99),
            f=random.random() * 10,
        )



import re
from functools import reduce
from itertools import chain
from typing import Union, Dict, List

import pandas as pd
import numpy as np

from .common import *

DataFrameType = Union[pd.DataFrame, Dict[str, pd.DataFrame], List[Dict[str, pd.DataFrame]]]


# Serialization helper functions
# -------------------------------

def _serializer(series) -> pd.DataFrame:
    df = pd.DataFrame(series.get('values', []), columns=series['columns'])
    if 'time' not in df.columns:
        return df
    df: pd.DataFrame = df.set_index(pd.to_datetime(df['time'])).drop('time', axis=1)
    df.index = df.index.tz_localize('UTC')
    df.index.name = None
    if 'tags' in series:
        for k, v in series['tags'].items():
            df[k] = v
    if 'name' in series:
        df.name = series['name']
    return df


def _get_name(series):
    tags = [f'{k}={v}' for k, v in series.get('tags', {}).items()]
    return ','.join(filter(None, [series.get('name'), *tags])) or None


def _drop_zero_index(df):
    if isinstance(df.index, pd.DatetimeIndex):
        if all(i.value == 0 for i in df.index):
            return df.reset_index(drop=True)
    return df


def parse(resp) -> DataFrameType:
    """Makes a dictionary of DataFrames from a response object"""
    statements = []
    for statement in resp['results']:
        series = {}
        for s in statement.get('series', []):
            series[_get_name(s)] = _drop_zero_index(_serializer(s))
        statements.append(series)

    if len(statements) == 1:
        series: dict = statements[0]
        if len(series) == 1:
            return list(series.values())[0]  # DataFrame
        else:
            return series  # dict
    return statements  # list


# Parsing helper functions
# -------------------------

def _itertuples(df):
    """Custom implementation of ``DataFrame.itertuples`` that
    returns plain tuples instead of namedtuples. About 50% faster.
    """
    cols = [df.iloc[:, k] for k in range(len(df.columns))]
    return zip(df.index, *cols)


def _replace(df):
    obj_cols = {k for k, v in dict(df.dtypes).items() if v is np.dtype('O')}
    other_cols = set(df.columns) - obj_cols
    obj_nans = (f'{k}="nan"' for k in obj_cols)
    other_nans = (f'{k}=nani?' for k in other_cols)
    replacements = [
        ('|'.join(chain(obj_nans, other_nans)), ''),
        (',{2,}', ','),
        ('|'.join([', ,', ', ', ' ,']), ' '),
    ]
    return replacements


def serialize(df, measurement, tag_columns=None, **extra_tags) -> bytes:
    """Converts a Pandas DataFrame into line protocol format"""
    # Pre-processing
    if measurement is None:
        raise ValueError("Missing 'measurement'")
    if not isinstance(df.index, pd.DatetimeIndex):
        raise ValueError('DataFrame index is not DatetimeIndex')
    tag_columns = set(tag_columns or [])
    isnull = df.isnull().any(axis=1)

    # Make parser function
    tags = []
    fields = []
    for k, v in extra_tags.items():
        tags.append(f"{k}={escape(v, key_escape)}")
    for i, (k, v) in enumerate(df.dtypes.items()):
        k = k.translate(key_escape)
        if k in tag_columns:
            tags.append(f"{k}={{p[{i+1}]}}")
        elif issubclass(v.type, np.integer):
            fields.append(f"{k}={{p[{i+1}]}}i")
        elif issubclass(v.type, (np.float, np.bool_)):
            fields.append(f"{k}={{p[{i+1}]}}")
        else:
            # String escaping is skipped for performance reasons
            # Strings containing double-quotes can cause strange write errors
            # and should be sanitized by the user.
            # e.g., df[k] = df[k].astype('str').str.translate(str_escape)
            fields.append(f"{k}=\"{{p[{i+1}]}}\"")
    fmt = (f'{measurement}', f'{"," if tags else ""}', ','.join(tags),
           ' ', ','.join(fields), ' {p[0].value}')
    f = eval("lambda p: f'{}'".format(''.join(fmt)))

    # Map/concat
    if isnull.any():
        lp = map(f, _itertuples(df[~isnull]))
        rep = _replace(df)
        lp_nan = (reduce(lambda a, b: re.sub(*b, a), rep, f(p))
                  for p in _itertuples(df[isnull]))
        return '\n'.join(chain(lp, lp_nan)).encode('utf-8')
    else:
        return '\n'.join(map(f, _itertuples(df))).encode('utf-8')
